/*
 * Copyright (c) 2000 Martin Jolicoeur <martinj@visuaide.com>
 * Portions Copyright (c) Independant JPEG group (ijg)
 * Portions Copyright (c) Greg Haerr <greg@censoft.com>
 *
 * Image mid-level loading routines
 *
 * These routines do the necessary range checking, clipping, and cursor
 * overwriting checks, and then call the lower level device dependent
 * routines to actually do the drawing.  The lower level routines are
 * only called when it is known that all the pixels to be drawn are
 * within the device area and are visible.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include "device.h"

#if defined(HAVE_FILEIO) && defined(HAVE_JPEG_SUPPORT)
#include "jpeglib.h"
#endif

#if 0
static void print_image(IMAGEHDR*);
#endif

/* BMP stuff */

#define PIX2BYTES(n)	(((n)+7)/8)

#define BI_RGB		0L
#define BI_RLE8		1L
#define BI_RLE4		2L
#define BI_BITFIELDS	3L

typedef unsigned char	BYTE;
typedef unsigned short	WORD;
typedef unsigned long	DWORD;
typedef long		LONG;

typedef struct {
	/* BITMAPFILEHEADER*/
	BYTE	bfType[2];
	DWORD	bfSize;
	WORD	bfReserved1;
	WORD	bfReserved2;
	DWORD	bfOffBits;
} BMPFILEHEAD;

#define FILEHEADSIZE 14

/* windows style*/
typedef struct {
	/* BITMAPINFOHEADER*/
	DWORD	BiSize;
	DWORD	BiWidth;
	DWORD	BiHeight;
	WORD	BiPlanes;
	WORD	BiBitCount;
	DWORD	BiCompression;
	DWORD	BiSizeImage;
	DWORD	BiXpelsPerMeter;
	DWORD	BiYpelsPerMeter;
	DWORD	BiClrUsed;
	DWORD	BiClrImportant;
} BMPINFOHEAD;

#define INFOHEADSIZE 40

/* os/2 style*/
typedef struct {
	/* BITMAPCOREHEADER*/
	DWORD	bcSize;
	WORD	bcWidth;
	WORD	bcHeight;
	WORD	bcPlanes;
	WORD	bcBitCount;
} BMPCOREHEAD;

#define COREHEADSIZE 12

static int	DecodeRLE8(UCHAR *buf,FILE *fp);
static int	DecodeRLE4(UCHAR *buf,FILE *fp);
static void	put4(int b);

/*
 * JPEG decompression routine
 *
 * Loads a jpeg from a file at runtime using libjpeg (http://www.ijg.org)
 *
 * destx, desty : image coordinates
 * width, height: maximum width & height to display (or IMAGE_MAX_SIZE to display all)
 * fast_grayscale: will output grayscale on grayscale display (faster)
 * filename: name of file to display
 *
 */
void GdJPEG (	PSD psd, COORD destx, COORD desty, COORD width, COORD height,
		BOOL fast_grayscale, char* filename)
{
#if defined(HAVE_FILEIO) && defined(HAVE_JPEG_SUPPORT)
	COORD minx, maxx;
	int clip;
	int k, l;
	RGBENTRY palette[256];
	SCREENINFO si;
	FILE *infile;		/* source file */
	JSAMPARRAY buffer;	/* Output row buffer */
	
	/* This struct contains the JPEG decompression parameters and pointers to
	 * working space (which is allocated as needed by the JPEG library).
	 */
	struct jpeg_decompress_struct cinfo;

	struct jpeg_error_mgr jerr;

	if ((infile = fopen (filename, "rb")) == NULL) {
		fprintf (stderr, "can't open %s\n", filename);
		return;
	}

	/* Step 1: allocate and initialize JPEG decompression object */

	/* We set up the normal JPEG error routines. */
	cinfo.err = jpeg_std_error (&jerr);

	/* Now we can initialize the JPEG decompression object. */
	jpeg_create_decompress (&cinfo);

	/* Step 2: specify data source (eg, a file) */
	jpeg_stdio_src (&cinfo, infile);

	/* Step 3: read file parameters with jpeg_read_header() */
	jpeg_read_header (&cinfo, TRUE);

	/* Step 4: set parameters for decompression */

	cinfo.out_color_space = fast_grayscale? JCS_GRAYSCALE: JCS_RGB;

	/* Get system information */
	GdGetScreenInfo(&scrdev, &si);

	/* Support only palettized for now ... */
	if (si.pixtype == PF_PALETTE)
	{
		cinfo.quantize_colors = TRUE;

		if (!fast_grayscale)
		{
			/* Get system palette */
			cinfo.actual_number_of_colors = GdGetPalette(0,si.ncolors,palette);
	
			/* Allocate jpeg colormap space */
			cinfo.colormap = (*cinfo.mem->alloc_sarray)((j_common_ptr) &cinfo, JPOOL_IMAGE,
			       	(JDIMENSION)cinfo.actual_number_of_colors, (JDIMENSION) 3);

			/* Set colormap from system palette */
			for(l = 0; l < cinfo.actual_number_of_colors; l++)
			{
				cinfo.colormap[0][l] = palette[l].r;
				cinfo.colormap[1][l] = palette[l].g;
				cinfo.colormap[2][l] = palette[l].b;
			}
		}
		else
		{
			/* Grayscale output asked */
			cinfo.out_color_space = JCS_GRAYSCALE;
			cinfo.desired_number_of_colors = si.ncolors;
		}
	}

	/* Step 5: Start decompressor */
	jpeg_start_decompress (&cinfo);

	/* Make a one-row-high sample array that will go away when done with image */
	buffer = (*cinfo.mem->alloc_sarray)
		((j_common_ptr) &cinfo, JPOOL_IMAGE,
		cinfo.output_width * cinfo.output_components, 1);

	if ((height > cinfo.output_height) || (height < 0)) {
		height = cinfo.output_height;
	}

	if ((width > cinfo.output_width) || (width < 0)) {
		width = cinfo.output_width;
	}

	/* determine if entire image is clipped out, save clipresult for later*/
	clip = GdClipArea(destx, desty, destx + width - 1, desty + height - 1);
	if(clip == CLIP_INVISIBLE)
		return;

	minx = destx;
	maxx = destx + width - 1;

	/* Step 6: while (scan lines remain to be read) */
	/*           jpeg_read_scanlines(...); */
	while (height > 0) {

		jpeg_read_scanlines (&cinfo, buffer, 1);

		k = 0;
		while (destx++ != maxx)
		{
			if(clip == CLIP_VISIBLE || GdClipPoint(destx, desty)) {
				psd->DrawPixel(psd, destx, desty, buffer[0][k++]);
			}
		}

		destx = minx;
		desty++;
		height--;
	}

	/*
	 * If we didn't output every scanlines, skip that step,
	 * otherwise it will generate an error and exit (see libjpeg.doc)
	 */
	if (cinfo.output_scanline == cinfo.output_height)
	{
		/* Step 7: Finish decompression */
		jpeg_finish_decompress (&cinfo);
	}

	/* Step 8: Release JPEG decompression object */
	/* This is an important step since it will release a good deal of memory. */
	jpeg_destroy_decompress (&cinfo);

	fclose (infile);

	/* At this point you may want to check to see whether any corrupt-data
	 * warnings occurred (test whether jerr.pub.num_warnings is nonzero).
	 */

	GdFixCursor();

	/* And we're done! */
#endif
}

/*
 * SOME FINE POINTS: (from libjpeg)
 *
 * In the above code, we ignored the return value of jpeg_read_scanlines,
 * which is the number of scanlines actually read.  We could get away with
 * this because we asked for only one line at a time and we weren't using
 * a suspending data source.  See libjpeg.doc for more info.
 *
 * We cheated a bit by calling alloc_sarray() after jpeg_start_decompress();
 * we should have done it beforehand to ensure that the space would be
 * counted against the JPEG max_memory setting.  In some systems the above
 * code would risk an out-of-memory error.  However, in general we don't
 * know the output image dimensions before jpeg_start_decompress(), unless we
 * call jpeg_calc_output_dimensions().  See libjpeg.doc for more about this.
 *
 * Scanlines are returned in the same order as they appear in the JPEG file,
 * which is standardly top-to-bottom.  If you must emit data bottom-to-top,
 * you can use one of the virtual arrays provided by the JPEG memory manager
 * to invert the data.  See wrbmp.c for an example.
 *
 * As with compression, some operating modes may require temporary files.
 * On some systems you may need to set up a signal handler to ensure that
 * temporary files are deleted if the program is interrupted.  See libjpeg.doc.
 */

/*
 * BMP decoding routine
 *
 * Loads a .bmp from a file at runtime
 *
 * destx, desty : image coordinates
 * width, height: maximum width & height to display (or IMAGE_MAX_SIZE to display all)
 * filename: name of file to display
 *
 */
void GdBMP (	PSD psd, COORD destx, COORD desty, COORD width, COORD height,
		char* filename)
{
#if defined(HAVE_FILEIO) && defined(HAVE_BMP_SUPPORT)
	UCHAR		bitvalue = 0;
	int		bitcount;
	PIXELVAL	pixel;
	COORD		yoff;
	PIXELVAL	convtable[256];
	COORD		minx, maxx;
	int		clip;
	FILE		*fp;
	BMPFILEHEAD	bmpf;
	BMPINFOHEAD	bmpi;
	BMPCOREHEAD	bmpc;
	int		i, compression, linesize;
	IMAGEHDR 	image;
	RGBENTRY 	palette[256];
	UCHAR		*imagebits;
	int		headsize;
	UCHAR 		headbuffer[INFOHEADSIZE];
	COLORVAL 	cr;

	/* Open file */
	if ((fp = fopen(filename, "rb")) == NULL)
		return;

	/* read BMP file header*/
	if(fread(&headbuffer, 1, FILEHEADSIZE, fp) != FILEHEADSIZE)
		goto err;

	bmpf.bfType[0] = headbuffer[0];
	bmpf.bfType[1] = headbuffer[1];
	bmpf.bfSize = *(DWORD*)&headbuffer[2];
	bmpf.bfOffBits = *(DWORD*)&headbuffer[10];

	/* Is it really a bmp file ? */
	if (*(WORD*)&bmpf.bfType[0] != 0x4D42) /* 'BM' */
		goto err;

	/* Read remaining header size */
	if (fread(&headsize, 1, sizeof(DWORD), fp) != sizeof(DWORD))
		goto err;

	/* might be windows or os/2 header */
	if(headsize == COREHEADSIZE) {

		/* read os/2 header */
		if(fread(&headbuffer, 1, COREHEADSIZE-sizeof(DWORD), fp) != COREHEADSIZE-sizeof(DWORD))
			goto err;

		/* Get data */
		bmpc.bcWidth = *(WORD*)&headbuffer[0];
		bmpc.bcHeight = *(WORD*)&headbuffer[2];
		bmpc.bcPlanes = *(WORD*)&headbuffer[4];
		bmpc.bcBitCount = *(WORD*)&headbuffer[6];
		
		image.width = (int)bmpc.bcWidth;
		image.height = (int)bmpc.bcHeight;
		image.bpp = bmpc.bcBitCount;
		image.palsize = 1 << image.bpp;
		compression = BI_RGB;
	} else {
		/* read windows header */
		if(fread(&headbuffer, 1, INFOHEADSIZE-sizeof(DWORD), fp) != INFOHEADSIZE-sizeof(DWORD))
			goto err;

		/* Get data */
		bmpi.BiWidth = *(DWORD*)&headbuffer[0];
		bmpi.BiHeight = *(DWORD*)&headbuffer[4];
		bmpi.BiPlanes = *(WORD*)&headbuffer[8];
		bmpi.BiBitCount = *(WORD*)&headbuffer[10];
		bmpi.BiCompression = *(DWORD*)&headbuffer[12];
		bmpi.BiSizeImage = *(DWORD*)&headbuffer[16];
		bmpi.BiXpelsPerMeter = *(DWORD*)&headbuffer[20];
		bmpi.BiYpelsPerMeter = *(DWORD*)&headbuffer[24];
		bmpi.BiClrUsed = *(DWORD*)&headbuffer[28];
		bmpi.BiClrImportant = *(DWORD*)&headbuffer[32];

		image.width = (int)bmpi.BiWidth;
		image.height = (int)bmpi.BiHeight;
		image.bpp = bmpi.BiBitCount;
		image.palsize = (int)bmpi.BiClrUsed;
		if(image.palsize == 0)
			image.palsize = 1 << image.bpp;
		compression = bmpi.BiCompression;
	}

	/* currently only decode palettized bitmaps*/
	if(image.bpp > 8) {
		fprintf(stderr, "Error: bits per pixel > 8\n");
		goto err;
	}

	/* compute image line size */
	if(image.bpp == 1)
		linesize = PIX2BYTES(image.width);
	else if(image.bpp <= 4)
		linesize = PIX2BYTES(image.width<<2);
	else if(image.bpp <= 8)
		linesize = image.width;
	else if(image.bpp <= 16)
		linesize = image.width * 2;
	else if(image.bpp <= 24)
		linesize = image.width * 3;
	else
		linesize = image.width * 4;

	linesize = (linesize+3) & ~3;

	/* Allocate image */
	if( (image.imagebits = malloc(linesize)) == NULL)
		goto err;

	/* get colormap*/
	if(image.bpp <= 8) {
		for(i=0; i<image.palsize; i++) {
			palette[i].b = fgetc(fp);
			palette[i].g = fgetc(fp);
			palette[i].r = fgetc(fp);
			if(headsize != COREHEADSIZE)
				fgetc(fp);
		}
	}

	image.compression = BI_RLE8;
	image.planes = 1;
	image.palette = &palette[0];

	imagebits = image.imagebits;

	/*
	 * Build a palette index conversion table.
	 */
	for(i=0; i<image.palsize; ++i) {
		cr = GETPALENTRY(image.palette, i);
		convtable[i] = GdFindColor(cr);
	}

	/* decode image data*/
	fseek(fp, bmpf.bfOffBits, SEEK_SET);

	if ((height > image.height) || (height < 0)) {
		height = image.height;
	}

	if ((width > image.width) || (width < 0)) {
		width = image.width;
	}

	/* determine if entire image is clipped out, save clipresult for later*/
	clip = GdClipArea(destx, desty, destx + width - 1, desty + height - 1);
	if(clip == CLIP_INVISIBLE)
		return;

	minx = destx;
	maxx = destx + width - 1;

	if(image.compression & 01) {
		/* bottom-up dibs*/
		desty += height - 1;
		yoff = -1;

		/* Eat extra bytes (bottom of image) 'till what we want */
		while (image.height > height)
		{
			/* Get row data from file */
			if(compression == BI_RLE8) {
				if(!DecodeRLE8(imagebits, fp))
					break;
			} else if(compression == BI_RLE4) {
				if(!DecodeRLE4(imagebits, fp))
					break;
			} else {
				if(fread(imagebits, 1, linesize, fp) != (size_t)(linesize))
					goto err;
			}

			image.height--;
		}
	} else
		yoff = 1;

	bitcount = 0;

	/* For every row ... */
	while (height > 0)
	{
		/* Get row data from file */
		if(compression == BI_RLE8) {
			if(!DecodeRLE8(imagebits, fp))
				break;
		} else if(compression == BI_RLE4) {
			if(!DecodeRLE4(imagebits, fp))
					break;
		} else {
			if(fread(imagebits, 1, linesize, fp) != (size_t)(linesize))
				goto err;
		}

		/* Output a row */
		while (destx++ != maxx)
		{
			if (bitcount <= 0) {
				bitcount = sizeof(UCHAR) * 8;
				bitvalue = *imagebits++;
			}

			switch (image.bpp) {
			default:
			case 8:
				pixel = bitvalue;
				bitcount = 0;
				break;
			case 4:
				pixel = (bitvalue & 0xf0) >> 4;
				bitvalue <<= 4;
				bitcount -= 4;
				break;
			case 1:
				pixel = (bitvalue & 0x80)? 1: 0;
				bitvalue <<= 1;
				--bitcount;
				break;
			}

			if(clip == CLIP_VISIBLE || GdClipPoint(destx, desty)) {
				psd->DrawPixel(psd, destx, desty, convtable[pixel]);
			}

		}

		destx = minx;
		desty += yoff;
		height--;
		bitcount = 0;
		imagebits = image.imagebits;
	}

	GdFixCursor();

err:
	if(image.imagebits)
		free(image.imagebits);
	fclose(fp);
	return;
#endif
}

/*
 * Decode one line of RLE8, return 0 when done with all bitmap data
 */
static int
DecodeRLE8(UCHAR *buf,FILE *fp)
{
	int		c, n;
	UCHAR *	p = buf;

	for( ;;) {
		switch( n = fgetc( fp)) {
		case EOF:
			return( 0);
		case 0:					/* 0 = escape*/
			switch( n = fgetc( fp)) {
			case 0: 			/* 0 0 = end of current scan line*/
				return( 1);
			case 1:				/* 0 1 = end of data*/
				return( 1);
			case 2:				/* 0 2 xx yy delta mode - NOT SUPPORTED*/
				(void)fgetc( fp);
				(void)fgetc( fp);
				continue;
			default:			/* 0 3..255 xx nn uncompressed data*/
				for( c=0; c<n; c++)
					*p++ = fgetc( fp);
				if( n & 1)
					(void)fgetc( fp);
				continue;
			}
		default:
			c = fgetc( fp);
			while( n--)
				*p++ = c;
			continue;
		}
	}
}

/*
 * Decode one line of RLE4, return 0 when done with all bitmap data
 */
static UCHAR *p;
static int	once;

static void
put4(int b)
{
	static int	last;

	last = (last << 4) | b;
	if( ++once == 2) {
		*p++ = last;
		once = 0;
	}
}
	
static int
DecodeRLE4(UCHAR *buf,FILE *fp)
{
	int		c, n, c1, c2;

	p = buf;
	once = 0;
	c1 = 0;

	for( ;;) {
		switch( n = fgetc( fp)) {
		case EOF:
			return( 0);
		case 0:					/* 0 = escape*/
			switch( n = fgetc( fp)) {
			case 0: 			/* 0 0 = end of current scan line*/
				if( once)
					put4( 0);
				return( 1);
			case 1:				/* 0 1 = end of data*/
				if( once)
					put4( 0);
				return( 1);
			case 2:				/* 0 2 xx yy delta mode - NOT SUPPORTED*/
				(void)fgetc( fp);
				(void)fgetc( fp);
				continue;
			default:			/* 0 3..255 xx nn uncompressed data*/
				c2 = (n+3) & ~3;
				for( c=0; c<c2; c++) {
					if( (c & 1) == 0)
							c1 = fgetc( fp);
					if( c < n)
							put4( (c1 >> 4) & 0x0f);
					c1 <<= 4;
				}
				continue;
			}
		default:
			c = fgetc( fp);
			c1 = (c >> 4) & 0x0f;
			c2 = c & 0x0f;
			for( c=0; c<n; c++)
				put4( (c&1)? c2: c1);
			continue;
		}
	}
}

#if 0
void print_image(IMAGEHDR* image)
{
	int i;

	printf("Image:\n\n");
	printf("height: %d\n", image->height);
	printf("width: %d\n", image->width);
	printf("planes: %d\n", image->planes);
	printf("bpp: %d\n", image->bpp);
	printf("compression: %d\n", image->compression);
	printf("palsize: %d\n", image->palsize);

	for (i=0;i<image->palsize;i++)
		printf("palette: %d, %d, %d\n", image->palette[i].r, image->palette[i].g, image->palette[i].b);

	for(i=0;i<(image->width*image->height);i++)
		printf("imagebits: %d\n", image->imagebits[i]);
}
#endif
